@@ -160,7 +160,7 @@ class Input(object):
         return iter(self.read())
 
     def get_socket(self):
-        return self.rfile._sock.dup()
+        return self.rfile._sock
 
 
 class HttpProtocol(BaseHTTPServer.BaseHTTPRequestHandler):
@@ -402,12 +402,10 @@ class HttpProtocol(BaseHTTPServer.BaseHTTPRequestHandler):
         env['REQUEST_METHOD'] = self.command
         env['SCRIPT_NAME'] = ''
 
-        if '?' in self.path:
-            path, query = self.path.split('?', 1)
-        else:
-            path, query = self.path, ''
-        env['PATH_INFO'] = urllib.unquote(path)
-        env['QUERY_STRING'] = query
+        pq = self.path.split('?', 1)
+        env['PATH_INFO'] = urllib.unquote(pq[0])
+        if len(pq) > 1:
+            env['QUERY_STRING'] = pq[1]
 
         if self.headers.typeheader is None:
             env['CONTENT_TYPE'] = self.headers.type
@@ -419,7 +417,7 @@ class HttpProtocol(BaseHTTPServer.BaseHTTPRequestHandler):
             env['CONTENT_LENGTH'] = length
         env['SERVER_PROTOCOL'] = 'HTTP/1.0'
 
-        host, port = self.request.getsockname()
+        host, port = self.request.getsockname()[:2]
         env['SERVER_NAME'] = host
         env['SERVER_PORT'] = str(port)
         env['REMOTE_ADDR'] = self.client_address[0]
@@ -498,6 +496,10 @@ class Server(BaseHTTPServer.HTTPServer):
             'wsgi.run_once': False,
             'wsgi.url_scheme': 'http',
         }
+        # detect secure socket
+        if hasattr(self.socket, 'do_handshake'):
+            d['wsgi.url_scheme'] = 'https'
+            d['HTTPS'] = 'on'
         if self.environ is not None:
             d.update(self.environ)
         return d
@@ -547,7 +549,7 @@ def server(sock, site,
     :param log_x_forwarded_for: If True (the default), logs the contents of the x-forwarded-for header in addition to the actual client ip address in the 'client_ip' field of the log line.
     :param custom_pool: A custom GreenPool instance which is used to spawn client green threads.  If this is supplied, max_size is ignored.
     :param keepalive: If set to False, disables keepalives on the server; all connections will be closed after serving one request.
-    :param log_format: A python format string that is used as the template to generate log lines.  The following values can be formatted into it: client_ip, date_time, request_line, status_code, body_length, wall_seconds.  Look the default for an example of how to use this.
+    :param log_format: A python format string that is used as the template to generate log lines.  The following values can be formatted into it: client_ip, date_time, request_line, status_code, body_length, wall_seconds.  The default is a good example of how to use it.
     """
     serv = Server(sock, sock.getsockname(),
                   site, log,
@@ -567,7 +569,7 @@ def server(sock, site,
     else:
         pool = greenpool.GreenPool(max_size)
     try:
-        host, port = sock.getsockname()
+        host, port = sock.getsockname()[:2]
         port = ':%s' % (port, )
         if hasattr(sock, 'do_handshake'):
             scheme = 'https'
